from setuptools import setup, find_packages

setup(
    name='envd_e2e_conda_test',
    version='1.0.0',
    url='https://github.com/mypackage.git',
    author='Author Name',
    author_email='author@gmail.com',
    description='Description of my package',
    packages=find_packages(),
)